/* $Id: rt_obj.c,v 1.3 2008/09/24 15:33:43 kiesling Exp $ */

/*
  This file is part of ctalk.
  Copyright  2005-2008  Robert Kiesling, ctalk@ctalklang.org.
  Permission is granted to copy this software provided that this copyright
  notice is included in all source code modules.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software Foundation, 
  Inc., 51 Franklin St., Fifth Floor, Boston, MA 02110-1301 USA.
*/

#include <stdio.h>
#include <string.h>
#include "object.h"
#include "message.h"
#include "ctalk.h"


extern OBJECT *__ctalk_dictionary;
extern OBJECT *__ctalk_last_object;

I_PASS interpreter_pass;             /* Declared in rtinfo.c.   */
extern RT_INFO *__call_stack[MAXARGS+1];
extern int __call_stack_ptr;

static METHOD *obj_method_by_fn (const OBJECT *__class_obj, OBJECT *(*fn)()) {

  static METHOD *__m;

  for (__m = __class_obj -> instance_methods; __m; __m = __m -> next)
    if (__m -> cfunc == fn)
      return __m;

  if (__class_obj -> __o_superclass) {
    return obj_method_by_fn (__class_obj -> __o_superclass, fn);
  }

  return NULL;
}

/* 
 *   TO DO - For now, if the class name is NULL, simply return 
 *   the first object that matches the name.  Add a 
 *   contextualization facility that allows the language to 
 *   qualify an ambiguous object with its class name.
 *
 *   NOTE - The return object should be checked in cases where,
 *   "name," refers to a method parameter name.
 *
 *   FIXME! -
 *     1. If, "name," refers to a parameter, also check local
 *        objects if a parameter is matched.
 */
OBJECT *__ctalk_get_object (const char *__name, const char *__classname) {

  OBJECT *__o,
    *__rcvr_obj,
    *(*parent_method_fn)();
  METHOD *__m;
  RT_FN *rt_fn;
  int i;

  /*
   *  Check for "self" keyword and parameter replacements.
   */

  if (!strcmp (__name, "self")) {
    __o = __ctalk_self_internal ();
    if (__classname) {
      if (!strcmp (__o -> __o_classname, __classname))
	return __o;
      else
	return NULL;
    } else {
      return __o;
    }
  } else {
    if (!strncmp (__name, "__ctalk_arg_internal", 
		  strlen ("__ctalk_arg_internal"))) {
      int arg_n;
      sscanf (__name, "__ctalk_arg_internal (%d)", &arg_n);
      return __ctalk_arg_internal (arg_n);
    }
  }

  /*
   *  If we're in a method, try to return a parameter, then local
   *  object.  The function should not need to check all of the
   *  receivers here, and __ctalk_getMethod* and __ctalk_findMethod*
   *  all call __ctalk_get_object (), so it's necessary to search for
   *  the method manually.
   *
   *  If the method is an inline method, first check the method's class
   *  if it is not NULL, because that is where local objects are stored.
   *
   */

  if (((__rcvr_obj = __ctalkRtMethodClassObject ()) != NULL) &&
      ((parent_method_fn = __ctalkRtGetMethodFn ()) != NULL)) {
    OBJECT *rcvr_class_obj;

    if (!strcmp (__rcvr_obj -> __o_classname, "Class")) {
      rcvr_class_obj = __rcvr_obj;
    } else {
      rcvr_class_obj = __rcvr_obj -> __o_class;
    }

    if ((__m = obj_method_by_fn (rcvr_class_obj, parent_method_fn)) != NULL) {

      for (i = 0; i < __m -> n_params; i++) {
	if (!strcmp (__m -> params[i] -> name, __name))
	  return __m -> args[__m -> n_params - i - 1];
      }


      if (__m -> local_objs) {
	for (__o = __m -> local_objs; __o; __o = __o -> next) {
	  if (__classname) {
	    if (!strcmp (__o -> __o_name, __name) && 
		!strcmp (__o -> __o_classname, __classname))
	      return __o;
	  } else {
	    if (!strcmp (__o -> __o_name, __name))
	      return __o;
	  }
	}
      }
    }
  }

  /*
   *  If not found, look up the parent method on the call stack if 
   *  the object is local to an inline method caller like map.
   */
  if (__call_stack_ptr < MAXARGS) {
    if (((__rcvr_obj = __call_stack[__call_stack_ptr+1]->rcvr_obj) != NULL) &&
	((parent_method_fn = __call_stack[__call_stack_ptr+1]->method_fn) 
	 != NULL)) {
      OBJECT *rcvr_class_obj;
      if (!strcmp (__rcvr_obj -> __o_classname, "Class")) {
	rcvr_class_obj = __rcvr_obj;
      } else {
	rcvr_class_obj = __rcvr_obj -> __o_class;
      }

      if (rcvr_class_obj) {
	if ((__m = obj_method_by_fn (rcvr_class_obj, parent_method_fn)) 
	    != NULL) {
	
	  for (i = 0; i < __m -> n_params; i++) {
	    if (!strcmp (__m -> params[i] -> name, __name))
	      return __m -> args[__m -> n_params - i - 1];
	  }

	  if (__m -> local_objs) {
	    for (__o = __m -> local_objs; __o; __o = __o -> next) {
	      if (__classname) {
		if (!strcmp (__o -> __o_name, __name) && 
		    !strcmp (__o -> __o_classname, __classname))
		  return __o;
	      } else {
		if (!strcmp (__o -> __o_name, __name))
		  return __o;
	      }
	    }
	  }
	}
      }

      /*
       *  Check for a global object.
       */

      if (__ctalk_dictionary) {
	for (__o = __ctalk_dictionary; ; __o = __o -> next) {
	  if (__classname) {
	    if (!strcmp (__o -> __o_name, __name) &&
		!strcmp (__o -> __o_classname, __classname))
	      return __o;
	  } else {
	    if (!strcmp (__o -> __o_name, __name))
	      return __o;
	  }
	  if (!__o || !__o -> next)
	    break;
	}
      }

      if (((rt_fn = get_fn ()) != NULL) ||
	  ((rt_fn = get_calling_fn ()) != NULL)) {
	for (__o = rt_fn -> local_objs; __o; __o = __o -> next) {
	  if (__classname) {
	    if (!strcmp (__name, __o -> __o_name) && 
		!strcmp (__classname, __o -> __o_classname))
	      return __o;
	  } else {
	    if (!strcmp (__name, __o -> __o_name))
	      return __o;
	  }
	}
      }

      /*
       * Also check for a class object.  TO DO - This is a duplication
       * of the code below, and should be factored out 
       * if this function gets more complicated.
       */

      if (__classname && !strcmp (__classname, "Class")) {
	return __ctalkGetClass (__name);
      }
      return NULL;
    }
  }

  /*
   *  If we're in a function, try to find the object there.
   */
  if (interpreter_pass == run_time_pass) {
    if ((rt_fn = get_fn ()) != NULL) {
      for (__o = rt_fn -> local_objs; __o; __o = __o -> next) {
 	if (__classname) {
 	  if (!strcmp (__name, __o -> __o_name) && 
 	      !strcmp (__classname, __o -> __o_classname))
 	    return __o;
 	} else {
 	  if (!strcmp (__name, __o -> __o_name))
 	    return __o;
 	}
      }
    }
  }

  if (__classname && !strcmp (__classname, "Class")) {
    return __ctalkGetClass (__name);
  } else {
    for (__o = __ctalk_dictionary; __o; __o = __o -> next) {
      if (__classname) {
	if (!strcmp (__o -> __o_name, __name) &&
	    !strcmp (__o -> __o_classname, __classname))
	  return __o;
      } else {
	if (!strcmp (__o -> __o_name, __name))
	  return __o;
      }
    }
  }

  /*
   *  Check for a class on its own.  Only CFunction class 
   *  methods need this at the moment.
   */
  if ((__o = __ctalkGetClass (__name)) != NULL)
    return __o;

  /*
   *  Finally, check for a class variable.
   */
  if ((__o = __ctalk_find_classvar (__name, __classname)) != NULL)
    return __o;

  return NULL;
}

/*
 *  This function should not need to check for, "self," and, "super,"
 *  because at this time it is only called from 
 *  __ctalkDeleteObjectInternal () and __ctalk_set_local ().
 */
int __ctalk_remove_object (OBJECT *o) {

  if (IS_CLASS_OBJECT(o)) {
    __ctalkRemoveClass (o);
  } else {
    if (o == __ctalk_last_object)
      __ctalk_last_object = __ctalk_last_object -> prev;
    if (o == __ctalk_dictionary)
      __ctalk_dictionary = __ctalk_dictionary -> next;
    if (o -> prev) o -> prev -> next = o -> next;
    if (o -> next) o -> next -> prev = o -> prev;
    o -> next = o -> prev = NULL;
  }

  return SUCCESS;
}

OBJECT *__ctalk_self_internal (void) {

  OBJECT *self_obj;

  self_obj = __ctalk_receiver_pop ();
  __ctalk_receiver_push (self_obj);
  return self_obj;
}

void __ctalk_set_global (char *name, char *classname) {

  OBJECT *obj;

  if ((obj = __ctalk_get_object (name, classname)) == NULL)
    _error ("__ctalk_set_global (): Undefined object %s (Class %s).\n", 
	    name, ((classname) ? classname : NULLSTR));

  obj -> scope = GLOBAL_VAR;

}

void __ctalk_set_local (OBJECT *obj, char *classname) {

  OBJECT *__method_rcvr_obj, 
    *(*fn)(),
    *o;
  METHOD *__m;
  RT_FN *rt_fn;

  obj -> scope = LOCAL_VAR;

  __objRefCntSet(OBJREF(obj), 1);

  /*
   *  If we're in a method, retrieve the method, and add the object
   *  there.
   */
  if (((__method_rcvr_obj = __ctalkRtReceiverObject ()) != NULL) &&
      ((fn = __ctalkRtGetMethodFn ()) != NULL) &&
      ((__m = 
	__ctalkFindInstanceMethodByFn (&__method_rcvr_obj, 
				       fn, 
				       TRUE)) != NULL)) {

    __ctalk_remove_object (obj);

    obj -> next = obj -> prev = NULL;

    if (__m -> local_objs) {

      for (o = __m -> local_objs; o -> next; o = o -> next)
	;
      o -> next = obj;
      obj -> prev = o;

    } else {
      __m -> local_objs = obj;
    }

    return;
  }

  /*
   *  If there's no receiver object, then we're in a C function.
   */

  if ((rt_fn = get_fn ()) == NULL) 
    return;

  __ctalk_remove_object (obj);

  obj -> next = obj -> prev = NULL;

  if (rt_fn -> local_objs) {

    for (o = rt_fn -> local_objs; o -> next; o = o -> next)
      ;
    o -> next = obj;
    obj -> prev = o;

  } else {
    rt_fn -> local_objs = obj;
  }
}

OBJECT *match_instancevar (OBJECT *varlist, OBJECT *var) {
  OBJECT *o, *p;
  for (o = varlist; o; o = o -> next) {
    if (o == var) {
      return o;
    } else {
      if (o -> instancevars)
	if ((p = match_instancevar (o -> instancevars, var)) != NULL)
	  return p;
    }
  }

  return NULL;
}

/*
 *  TO DO - Only used by __ctalkDeleteObjectInternal.  Should be 
 *  generalized.
 */
OBJECT *__ctalkInstanceVariableObject (OBJECT *var) {

  OBJECT *r, *o;

  for (o = __ctalk_dictionary; o; o = o -> next) {
    if ((r = match_instancevar (o -> instancevars, var)) != NULL)
      return o;
  }
  return NULL;
}

/*
 *  Like __ctalkInstanceVariableObject, but checks for 
 *  instance variable membership of objects on the receiver
 *  stack.
 */

OBJECT *__ctalkInstanceVariableReceiver (OBJECT *possible_var) {

  int i;
  OBJECT *rcvr, *v;

  for (i = __ctalkGetReceiverPtr () + 1; i <= MAXARGS; i++) {
    rcvr = __ctalk_receivers_at (i);
    for (v = rcvr -> instancevars; v; v = v -> next) {
      if (v == possible_var)
	return rcvr;
    }
  }
  return possible_var;
}

OBJECT *__ctalkCopyObject (OBJREF_T src_ptr, OBJREF_T dest_ptr) {
  OBJECT *p, *p_var, *p_var2, *new_var;

  *dest_ptr = __ctalkCreateObject ((*src_ptr) -> __o_name, 
				   (*src_ptr) -> __o_classname, 
				   (((*src_ptr) -> __o_superclassname) ? 
				    (*src_ptr) -> __o_superclassname : NULL),
				   (*src_ptr) -> scope);
  __ctalkSetObjectValue (*dest_ptr, (*src_ptr) -> __o_value);

  for (p_var = (*src_ptr) -> instancevars; p_var; p_var = p_var -> next) {
    /*
     *  The, "value," instance variable must already exist for 
     *  __ctalkAddInstanceVariable to be called, so add, "value," 
     *  manually here.
     */
    if (!strcmp (p_var -> __o_name, "value")) {
      if ((*dest_ptr) -> instancevars) {
	for (p_var2 = p -> instancevars; p_var2 && p_var2 -> next;
	     p_var2 = p_var2 -> next)
	  ;
	__ctalkCopyObject (OBJREF(p_var), OBJREF(new_var));
	p_var2 -> next = new_var;
	new_var -> prev = p_var2;
	if (p_var2 -> __o_p_obj)
	  new_var -> __o_p_obj = p_var2 -> __o_p_obj;
      } else {
	__ctalkCopyObject (OBJREF(p_var), 
			   OBJREF((*dest_ptr) -> instancevars));
	if (p_var -> __o_p_obj)
	  (*dest_ptr) -> instancevars -> __o_p_obj = (*dest_ptr);
      }
    } else {
      __ctalkCopyObject (OBJREF(p_var), OBJREF(new_var));
      __ctalkAddInstanceVariable (*dest_ptr, p_var -> __o_name, new_var);
    }
  }
    
  return *dest_ptr;
}

void __ctalkSetObjectScope (OBJECT *__o, int scope) {

  OBJECT *__var;

  for (__var = __o -> instancevars; __var; __var = __var -> next) 
    __ctalkSetObjectScope (__var, scope);
  
  __o -> scope = scope;
}
